"use strict";
/*!
 * Copyright (c) Microsoft Corporation and contributors. All rights reserved.
 * Licensed under the MIT License.
 */
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.connectDocument = void 0;
const protocol_definitions_1 = require("@fluidframework/protocol-definitions");
const server_services_client_1 = require("@fluidframework/server-services-client");
const server_services_core_1 = require("@fluidframework/server-services-core");
const server_services_telemetry_1 = require("@fluidframework/server-services-telemetry");
const json_stringify_safe_1 = __importDefault(require("json-stringify-safe"));
const utils_1 = require("../utils");
const utils_2 = require("./utils");
const trace_1 = require("./trace");
const protocol_1 = require("./protocol");
const throttleAndUsage_1 = require("./throttleAndUsage");
/**
 * Trace stages for the connect flow.
 */
var ConnectDocumentStage;
(function (ConnectDocumentStage) {
    /**
     * The connect document flow has started.
     */
    ConnectDocumentStage["ConnectDocumentStarted"] = "ConnectDocumentStarted";
    /**
     * Connection protocol versions have been parsed and validated.
     */
    ConnectDocumentStage["VersionsChecked"] = "VersionsChecked";
    /**
     * Connection throttling has been checked and updated.
     */
    ConnectDocumentStage["ThrottleChecked"] = "ThrottleChecked";
    /**
     * Authentication token has been validated.
     */
    ConnectDocumentStage["TokenVerified"] = "TokenVerified";
    /**
     * Socket rooms have been joined/subscribed to.
     */
    ConnectDocumentStage["RoomJoined"] = "RoomJoined";
    /**
     * Connected clients list has been retrieved from the client manager.
     */
    ConnectDocumentStage["ClientsRetrieved"] = "ClientsRetrieved";
    /**
     * Server-trusted client information has been compiled into a "message client" variable.
     */
    ConnectDocumentStage["MessageClientCreated"] = "MessageClientCreated";
    /**
     * Message client has been added to the client manager's connected clients list.
     */
    ConnectDocumentStage["MessageClientAdded"] = "MessageClientAdded";
    /**
     * A timer has been started to terminate the connection when the token expires.
     */
    ConnectDocumentStage["TokenExpirySet"] = "TokenExpirySet";
    /**
     * The orderer connection has been established (if client is a writer) and the
     * connected message response to send back to the client has been composed.
     */
    ConnectDocumentStage["MessageClientConnected"] = "MessageClientConnected";
    /**
     * Socket tracking for this connection has been started for token revocation purposes.
     */
    ConnectDocumentStage["SocketTrackerAppended"] = "SocketTrackerAppended";
    /**
     * Collaboration session tracking for this connection has been started for telemetry purposes.
     */
    ConnectDocumentStage["SessionTrackerAppended"] = "SessionTrackerAppended";
    /**
     * A listener has been set up to broadcast signals from external APIs to the room.
     */
    ConnectDocumentStage["SignalListenerSetUp"] = "SignalListenerSetUp";
    /**
     * The client has been successfully joined to the room and a client join notification has been
     * emitted to the room.
     */
    ConnectDocumentStage["JoinSignalEmitted"] = "JoinSignalEmitted";
})(ConnectDocumentStage || (ConnectDocumentStage = {}));
function composeConnectedMessage(messageSize, mode, serviceConfigurationBlockSize, serviceConfigurationMaxMessageSize, clientId, claims, version, clients) {
    const connectedMessage = {
        claims,
        clientId,
        existing: true,
        maxMessageSize: messageSize,
        mode,
        serviceConfiguration: {
            blockSize: serviceConfigurationBlockSize,
            maxMessageSize: serviceConfigurationMaxMessageSize,
        },
        initialClients: clients,
        initialMessages: [],
        initialSignals: [],
        supportedVersions: protocol_1.ProtocolVersions,
        supportedFeatures: {
            submit_signals_v2: true,
        },
        version,
    };
    return connectedMessage;
}
async function connectOrderer(socket, lumberjackProperties, lambdaDependencies, lambdaSettings, lambdaConnectionStateTrackers, tenantId, documentId, messageClient, startTime, clientId, claims, version, clients) {
    const { ordererManager, logger } = lambdaDependencies;
    const { numberOfMessagesPerTrace } = lambdaSettings;
    const { expirationTimer, connectionsMap } = lambdaConnectionStateTrackers;
    const connectDocumentOrdererConnectionMetric = server_services_telemetry_1.Lumberjack.newLumberMetric(server_services_telemetry_1.LumberEventName.ConnectDocumentOrdererConnection, lumberjackProperties);
    const orderer = await ordererManager.getOrderer(tenantId, documentId).catch(async (error) => {
        const errMsg = `Failed to get orderer manager. Error: ${(0, json_stringify_safe_1.default)(error, undefined, 2)}`;
        connectDocumentOrdererConnectionMetric.error("Failed to get orderer manager", error);
        throw (0, utils_2.handleServerErrorAndConvertToNetworkError)(logger, errMsg, documentId, tenantId, error);
    });
    const connection = await orderer
        .connect(socket, clientId, messageClient)
        .catch(async (error) => {
        const errMsg = `Failed to connect to orderer. Error: ${(0, json_stringify_safe_1.default)(error, undefined, 2)}`;
        connectDocumentOrdererConnectionMetric.error("Failed to connect to orderer", error);
        throw (0, utils_2.handleServerErrorAndConvertToNetworkError)(logger, errMsg, documentId, tenantId, error);
    });
    // Eventually we will send disconnect reason as headers to client.
    const connectionErrorListener = (error) => {
        const messageMetaData = (0, utils_2.getMessageMetadata)(connection.documentId, connection.tenantId);
        logger.error(`Disconnecting socket on connection error: ${(0, json_stringify_safe_1.default)(error, undefined, 2)}`, { messageMetaData });
        server_services_telemetry_1.Lumberjack.error(`Disconnecting socket on connection error`, (0, server_services_telemetry_1.getLumberBaseProperties)(connection.documentId, connection.tenantId), error);
        expirationTimer.clear();
        socket.disconnect(true);
    };
    connection.once("error", connectionErrorListener);
    let clientJoinMessageServerMetadata;
    if (server_services_core_1.DefaultServiceConfiguration.enableTraces && (0, trace_1.sampleMessages)(numberOfMessagesPerTrace)) {
        clientJoinMessageServerMetadata = {
            connectDocumentStartTime: startTime,
        };
    }
    connection.connect(clientJoinMessageServerMetadata).catch(async (error) => {
        const errMsg = `Failed to connect to the orderer connection. Error: ${(0, json_stringify_safe_1.default)(error, undefined, 2)}`;
        connectDocumentOrdererConnectionMetric.error("Failed to establish orderer connection", error);
        throw (0, utils_2.handleServerErrorAndConvertToNetworkError)(logger, errMsg, documentId, tenantId, error);
    });
    connectionsMap.set(clientId, connection);
    if (connectionsMap.size > 1) {
        server_services_telemetry_1.Lumberjack.info(`Same socket is having multiple connections, connection number=${connectionsMap.size}`, (0, server_services_telemetry_1.getLumberBaseProperties)(connection.documentId, connection.tenantId));
    }
    connectDocumentOrdererConnectionMetric.success("Successfully established orderer connection");
    const connectedMessage = composeConnectedMessage(connection.maxMessageSize, "write", connection.serviceConfiguration.blockSize, connection.serviceConfiguration.maxMessageSize, clientId, claims, version, clients);
    return {
        connectedMessage,
        disposeOrdererConnectionListener: () => {
            connection.off("error", connectionErrorListener);
        },
    };
}
function trackSocket(socket, tenantId, documentId, claims, { socketTracker }) {
    // Track socket and tokens for this connection
    if (socketTracker && claims.jti) {
        socketTracker.addSocketForToken((0, server_services_core_1.createCompositeTokenId)(tenantId, documentId, claims.jti), socket);
    }
}
function trackCollaborationSession(clientId, clientDetails, isWriteClient, tenantId, documentId, connectedClients, { collaborationSessionTracker }) {
    // Track the collaboration session for this connection
    if (collaborationSessionTracker) {
        const sessionClient = {
            clientId,
            joinedTime: clientDetails.timestamp ?? Date.now(),
            isWriteClient,
            isSummarizerClient: (0, utils_2.isSummarizer)(clientDetails.details),
        };
        collaborationSessionTracker
            .startClientSession(sessionClient, { documentId, tenantId }, connectedClients)
            .catch((error) => {
            server_services_telemetry_1.Lumberjack.error("Failed to update collaboration session tracker for new client", { tenantId, documentId }, error);
        });
    }
}
function checkThrottle(tenantId, { throttlers, logger }) {
    const throttleErrorPerCluster = (0, throttleAndUsage_1.checkThrottleAndUsage)(throttlers.connectionsPerCluster, (0, throttleAndUsage_1.getSocketConnectThrottleId)("connectDoc"), tenantId, logger);
    if (throttleErrorPerCluster) {
        // eslint-disable-next-line @typescript-eslint/no-throw-literal
        throw throttleErrorPerCluster;
    }
    const throttleErrorPerTenant = (0, throttleAndUsage_1.checkThrottleAndUsage)(throttlers.connectionsPerTenant, (0, throttleAndUsage_1.getSocketConnectThrottleId)(tenantId), tenantId, logger);
    if (throttleErrorPerTenant) {
        // eslint-disable-next-line @typescript-eslint/no-throw-literal
        throw throttleErrorPerTenant;
    }
}
async function checkToken(token, tenantId, documentId, { tenantManager, revokedTokenChecker, logger }) {
    if (!token) {
        throw new server_services_client_1.NetworkError(403, "Must provide an authorization token");
    }
    const claims = (0, server_services_client_1.validateTokenClaims)(token, documentId, tenantId);
    try {
        if (revokedTokenChecker && claims.jti) {
            const isTokenRevoked = await revokedTokenChecker.isTokenRevoked(claims.tenantId, claims.documentId, claims.jti);
            if (isTokenRevoked) {
                const error = (0, server_services_client_1.createFluidServiceNetworkError)(403, {
                    message: "Permission denied. Token has been revoked",
                    internalErrorCode: server_services_client_1.InternalErrorCode.TokenRevoked,
                    canRetry: false,
                    isFatal: true,
                });
                throw error;
            }
        }
        await tenantManager.verifyToken(claims.tenantId, token);
        return claims;
    }
    catch (error) {
        if ((0, server_services_client_1.isNetworkError)(error)) {
            throw error;
        }
        // We don't understand the error, so it is likely an internal service error.
        const errMsg = `Could not verify connect document token. Error: ${(0, json_stringify_safe_1.default)(error, undefined, 2)}`;
        throw (0, utils_2.handleServerErrorAndConvertToNetworkError)(logger, errMsg, claims.documentId, claims.tenantId, error);
    }
}
async function checkClusterDraining({ clusterDrainingChecker }, message, properties) {
    if (!clusterDrainingChecker) {
        return;
    }
    let clusterInDraining = false;
    try {
        clusterInDraining = await clusterDrainingChecker.isClusterDraining({
            tenantId: message.tenantId,
        });
    }
    catch (error) {
        server_services_telemetry_1.Lumberjack.error("Failed to get cluster draining status. Will allow requests to proceed.", properties, error);
        clusterInDraining = false;
    }
    if (clusterInDraining) {
        server_services_telemetry_1.Lumberjack.info("Reject connect document request because cluster is draining.", {
            ...properties,
            tenantId: message.tenantId,
        });
        const error = (0, server_services_client_1.createFluidServiceNetworkError)(503, {
            message: "Cluster is not available. Please retry later.",
            internalErrorCode: server_services_client_1.InternalErrorCode.ClusterDraining,
            retryAfterMs: server_services_core_1.clusterDrainingRetryTimeInMs,
        });
        throw error;
    }
}
async function joinRoomAndSubscribeToChannel(socket, tenantId, documentId, { logger }) {
    const clientId = (0, utils_1.generateClientId)();
    const room = {
        tenantId,
        documentId,
    };
    try {
        // Subscribe to channels.
        await Promise.all([
            socket.join((0, utils_2.getRoomId)(room)),
            socket.join((0, utils_2.getClientSpecificRoomId)(clientId)),
        ]);
        return [clientId, room];
    }
    catch (error) {
        const errMsg = `Could not subscribe to channels. Error: ${(0, json_stringify_safe_1.default)(error, undefined, 2)}`;
        throw (0, utils_2.handleServerErrorAndConvertToNetworkError)(logger, errMsg, documentId, tenantId, error);
    }
}
async function retrieveClients(tenantId, documentId, metricProperties, { clientManager, logger }, { maxNumberOfClientsPerDocument }) {
    const connectDocumentGetClientsMetric = server_services_telemetry_1.Lumberjack.newLumberMetric(server_services_telemetry_1.LumberEventName.ConnectDocumentGetClients, metricProperties);
    const clients = await clientManager
        .getClients(tenantId, documentId)
        .then((response) => {
        connectDocumentGetClientsMetric.success("Successfully got clients from client manager");
        return response;
    })
        .catch(async (error) => {
        const errMsg = `Failed to get clients. Error: ${(0, json_stringify_safe_1.default)(error, undefined, 2)}`;
        connectDocumentGetClientsMetric.error("Failed to get clients during connectDocument", error);
        throw (0, utils_2.handleServerErrorAndConvertToNetworkError)(logger, errMsg, documentId, tenantId, error);
    });
    if (clients.length > maxNumberOfClientsPerDocument) {
        throw new server_services_client_1.NetworkError(429, "Too Many Clients Connected to Document", true /* canRetry */, false /* isFatal */, 5 * 60 * 1000 /* retryAfterMs (5 min) */);
    }
    return clients;
}
function createMessageClient(mode, client, claims, room, clientId, connectedTimestamp, supportedFeatures, { connectionTimeMap, scopeMap, roomMap, clientMap, supportedFeaturesMap, }) {
    // Todo should all the client details come from the claims???
    // we are still trusting the users permissions and type here.
    const messageClient = {
        ...client,
        user: claims.user,
        mode: (0, utils_2.isWriter)(claims.scopes, mode) ? "write" : "read",
        scopes: claims.scopes,
    };
    // 1. Do not give SummaryWrite scope to clients that are not summarizers.
    // 2. Store connection timestamp for all clients but the summarizer.
    // Connection timestamp is used (inside socket disconnect event) to
    // calculate the client connection time (i.e. for billing).
    const isSummarizerClient = (0, utils_2.isSummarizer)(messageClient.details);
    if (!isSummarizerClient) {
        messageClient.scopes = claims.scopes.filter((scope) => scope !== protocol_definitions_1.ScopeType.SummaryWrite);
        connectionTimeMap.set(clientId, connectedTimestamp);
    }
    // Cache the scopes.
    scopeMap.set(clientId, messageClient.scopes);
    // Join the room to receive signals.
    roomMap.set(clientId, room);
    // Store the supported features for the client
    supportedFeaturesMap.set(clientId, supportedFeatures ?? {});
    // Store the client details.
    clientMap.set(clientId, messageClient);
    return messageClient;
}
async function addMessageClientToClientManager(tenantId, documentId, clientId, messageClient, metricProperties, { clientManager, logger }) {
    const connectDocumentAddClientMetric = server_services_telemetry_1.Lumberjack.newLumberMetric(server_services_telemetry_1.LumberEventName.ConnectDocumentAddClient, metricProperties);
    try {
        await clientManager.addClient(tenantId, documentId, clientId, messageClient);
        connectDocumentAddClientMetric.success("Successfully added client");
    }
    catch (error) {
        const errMsg = `Could not add client. Error: ${(0, json_stringify_safe_1.default)(error, undefined, 2)}`;
        connectDocumentAddClientMetric.error("Error adding client during connectDocument", error);
        throw (0, utils_2.handleServerErrorAndConvertToNetworkError)(logger, errMsg, documentId, tenantId, error);
    }
}
/**
 * Sets up a listener for broadcast signals from external API and broadcasts them to the room.
 *
 * @param socket - Websocket instance
 * @param room - Current room
 * @param lambdaDependencies - Lambda dependencies including collaborationSessionEventEmitter and logger
 * @returns Dispose function to remove the added listener
 */
function setUpSignalListenerForRoomBroadcasting(socket, room, { collaborationSessionEventEmitter, logger }) {
    const broadCastSignalListener = (broadcastSignal) => {
        const { signalRoom, signalContent } = broadcastSignal;
        // No-op if the room (collab session) that signal came in from is different
        // than the current room. We reuse websockets so there could be multiple rooms
        // that we are sending the signal to, and we don't want to do that.
        if (signalRoom.documentId === room.documentId && signalRoom.tenantId === room.tenantId) {
            try {
                const runtimeMessage = (0, utils_1.createRuntimeMessage)(signalContent);
                try {
                    socket.emitToRoom((0, utils_2.getRoomId)(signalRoom), "signal", runtimeMessage);
                }
                catch (error) {
                    const errorMsg = `Failed to broadcast signal from external API.`;
                    server_services_telemetry_1.Lumberjack.error(errorMsg, (0, server_services_telemetry_1.getLumberBaseProperties)(signalRoom.documentId, signalRoom.tenantId), error);
                }
            }
            catch (error) {
                const errorMsg = `broadcast-signal content body is malformed`;
                throw (0, utils_2.handleServerErrorAndConvertToNetworkError)(logger, errorMsg, room.documentId, room.tenantId, error);
            }
        }
    };
    collaborationSessionEventEmitter?.on("broadcastSignal", broadCastSignalListener);
    const disposeBroadcastSignalListener = () => {
        collaborationSessionEventEmitter?.off("broadcastSignal", broadCastSignalListener);
    };
    return disposeBroadcastSignalListener;
}
async function connectDocument(socket, lambdaDependencies, lambdaSettings, lambdaConnectionStateTrackers, message, properties, denyList) {
    const { isTokenExpiryEnabled, maxTokenLifetimeSec } = lambdaSettings;
    const { expirationTimer } = lambdaConnectionStateTrackers;
    const connectionTrace = new trace_1.StageTrace(ConnectDocumentStage.ConnectDocumentStarted);
    const startTime = Date.now();
    const connectMetric = server_services_telemetry_1.Lumberjack.newLumberMetric(server_services_telemetry_1.LumberEventName.ConnectDocument, properties);
    let tenantId = message.tenantId;
    let documentId = message.id;
    let uncaughtError;
    try {
        const [connectVersions, version] = (0, protocol_1.checkProtocolVersion)(message.versions);
        connectionTrace.stampStage(ConnectDocumentStage.VersionsChecked);
        checkThrottle(tenantId, lambdaDependencies);
        connectionTrace.stampStage(ConnectDocumentStage.ThrottleChecked);
        const claims = await checkToken(message.token, tenantId, documentId, lambdaDependencies);
        // check token validate tenantId/documentId for consistent, throw 403 if now.
        // Following change tenantId/documentId from claims, just in case future code changes that we can remember to use the ones from claim.
        tenantId = claims.tenantId;
        documentId = claims.documentId;
        connectionTrace.stampStage(ConnectDocumentStage.TokenVerified);
        if (denyList?.isTenantDenied(tenantId)) {
            server_services_telemetry_1.Lumberjack.error("Tenant is in the deny list", {
                ...properties,
                tenantId,
            });
            throw new server_services_client_1.NetworkError(500, `Unable to process request for tenant id: ${tenantId}`);
        }
        if (denyList?.isDocumentDenied(documentId)) {
            server_services_telemetry_1.Lumberjack.error("Document is in the deny list", {
                ...properties,
                tenantId,
                documentId,
            });
            throw new server_services_client_1.NetworkError(500, `Unable to process request for document id: ${documentId}`);
        }
        await checkClusterDraining(lambdaDependencies, message, properties);
        const [clientId, room] = await joinRoomAndSubscribeToChannel(socket, tenantId, documentId, lambdaDependencies);
        connectionTrace.stampStage(ConnectDocumentStage.RoomJoined);
        const subMetricProperties = {
            ...(0, server_services_telemetry_1.getLumberBaseProperties)(documentId, tenantId),
            [server_services_telemetry_1.CommonProperties.clientId]: clientId,
        };
        const clients = await retrieveClients(tenantId, documentId, subMetricProperties, lambdaDependencies, lambdaSettings);
        connectionTrace.stampStage(ConnectDocumentStage.ClientsRetrieved);
        const connectedTimestamp = Date.now();
        const messageClient = createMessageClient(message.mode, message.client, claims, room, clientId, connectedTimestamp, message.supportedFeatures, lambdaConnectionStateTrackers);
        connectionTrace.stampStage(ConnectDocumentStage.MessageClientCreated);
        await addMessageClientToClientManager(tenantId, documentId, clientId, messageClient, subMetricProperties, lambdaDependencies);
        connectionTrace.stampStage(ConnectDocumentStage.MessageClientAdded);
        if (isTokenExpiryEnabled) {
            const lifeTimeMSec = (0, server_services_client_1.validateTokenClaimsExpiration)(claims, maxTokenLifetimeSec);
            expirationTimer.set(lifeTimeMSec);
        }
        connectionTrace.stampStage(ConnectDocumentStage.TokenExpirySet);
        const isWriterClient = (0, utils_2.isWriter)(messageClient.scopes, message.mode);
        connectMetric.setProperty("IsWriterClient", isWriterClient);
        const { connectedMessage, disposeOrdererConnectionListener } = isWriterClient
            ? await connectOrderer(socket, subMetricProperties, lambdaDependencies, lambdaSettings, lambdaConnectionStateTrackers, tenantId, documentId, messageClient, startTime, clientId, claims, version, clients)
            : {
                connectedMessage: composeConnectedMessage(1024 /* messageSize */, "read", server_services_core_1.DefaultServiceConfiguration.blockSize, server_services_core_1.DefaultServiceConfiguration.maxMessageSize, clientId, claims, version, clients),
                disposeOrdererConnectionListener: () => { },
            };
        // back-compat: remove cast to any once new definition of IConnected comes through.
        connectedMessage.timestamp = connectedTimestamp;
        connectionTrace.stampStage(ConnectDocumentStage.MessageClientConnected);
        trackSocket(socket, tenantId, documentId, claims, lambdaDependencies);
        connectionTrace.stampStage(ConnectDocumentStage.SocketTrackerAppended);
        trackCollaborationSession(clientId, { ...messageClient, mode: message.mode }, isWriterClient, tenantId, documentId, clients, lambdaDependencies);
        const disposeSignalListenerForRoomBroadcasting = setUpSignalListenerForRoomBroadcasting(socket, room, lambdaDependencies);
        connectionTrace.stampStage(ConnectDocumentStage.SignalListenerSetUp);
        const result = {
            connection: connectedMessage,
            connectVersions,
            details: messageClient,
        };
        socket.emitToRoom((0, utils_2.getRoomId)(room), "signal", (0, utils_1.createRoomJoinMessage)(result.connection.clientId, result.details));
        connectionTrace.stampStage(ConnectDocumentStage.JoinSignalEmitted);
        connectMetric.setProperties({
            [server_services_telemetry_1.CommonProperties.clientId]: result.connection.clientId,
            [server_services_telemetry_1.CommonProperties.clientCount]: result.connection.initialClients.length + 1,
            [server_services_telemetry_1.CommonProperties.clientType]: result.details.details?.type,
        });
        return {
            connection: connectedMessage,
            connectVersions,
            details: messageClient,
            dispose: () => {
                disposeSignalListenerForRoomBroadcasting();
                disposeOrdererConnectionListener();
            },
        };
    }
    catch (error) {
        uncaughtError = error;
        throw error;
    }
    finally {
        connectMetric.setProperty("connectTrace", connectionTrace);
        if (uncaughtError) {
            if (uncaughtError.code !== undefined) {
                connectMetric.setProperty(server_services_telemetry_1.CommonProperties.errorCode, uncaughtError.code);
            }
            if (uncaughtError.internalErrorCode !== undefined) {
                connectMetric.setProperty(server_services_telemetry_1.CommonProperties.internalErrorCode, uncaughtError.internalErrorCode);
            }
            connectMetric.error(`Connect document failed`, uncaughtError);
        }
        else {
            connectMetric.success(`Connect document successful`);
        }
    }
}
exports.connectDocument = connectDocument;
//# sourceMappingURL=connect.js.map