"use strict";
/*!
 * Copyright (c) Microsoft Corporation and contributors. All rights reserved.
 * Licensed under the MIT License.
 */
Object.defineProperty(exports, "__esModule", { value: true });
exports.disconnectDocument = void 0;
const server_services_telemetry_1 = require("@fluidframework/server-services-telemetry");
const utils_1 = require("../utils");
const utils_2 = require("./utils");
const throttleAndUsage_1 = require("./throttleAndUsage");
/**
 * Disconnect all orderer connections and store connectivity time for each.
 */
function disconnectOrdererConnections({ ordererManager, logger, throttleAndUsageStorageManager }, { isClientConnectivityCountingEnabled }, { connectionTimeMap, connectionsMap, disconnectedOrdererConnections, }) {
    const promises = [];
    for (const [clientId, connection] of connectionsMap) {
        if (disconnectedOrdererConnections.has(clientId)) {
            // We already removed this clientId once. Skip it.
            continue;
        }
        const messageMetaData = (0, utils_2.getMessageMetadata)(connection.documentId, connection.tenantId);
        logger.info(`Disconnect of ${clientId}`, { messageMetaData });
        const lumberjackProperties = (0, server_services_telemetry_1.getLumberBaseProperties)(connection.documentId, connection.tenantId);
        server_services_telemetry_1.Lumberjack.info(`Disconnect of ${clientId}`, lumberjackProperties);
        connection
            .disconnect()
            .then(() => {
            // Keep track of disconnected clientIds so that we don't repeat the disconnect signal
            // for the same clientId if retrying when connectDocument completes after disconnectDocument.
            disconnectedOrdererConnections.add(clientId);
            ordererManager.removeOrderer(connection.tenantId, connection.documentId);
        })
            .catch((error) => {
            const errorMsg = `Failed to disconnect client ${clientId} from orderer connection.`;
            server_services_telemetry_1.Lumberjack.error(errorMsg, lumberjackProperties, error);
        });
        if (isClientConnectivityCountingEnabled && throttleAndUsageStorageManager) {
            const connectionTimestamp = connectionTimeMap.get(clientId);
            if (connectionTimestamp) {
                promises.push((0, throttleAndUsage_1.storeClientConnectivityTime)(clientId, connection.documentId, connection.tenantId, connectionTimestamp, throttleAndUsageStorageManager));
            }
        }
    }
    return promises;
}
/**
 * Remove clients from the client manager and send client leave notifications to the room.
 */
function removeClientAndSendNotifications(socket, { clientManager, logger, collaborationSessionTracker }, { roomMap, clientMap, connectionTimeMap, disconnectedClients, }) {
    const promises = [];
    for (const [clientId, room] of roomMap) {
        if (disconnectedClients.has(clientId)) {
            // We already removed this clientId once. Skip it.
            continue;
        }
        const messageMetaData = (0, utils_2.getMessageMetadata)(room.documentId, room.tenantId);
        logger.info(`Disconnect of ${clientId} from room`, { messageMetaData });
        server_services_telemetry_1.Lumberjack.info(`Disconnect of ${clientId} from room`, (0, server_services_telemetry_1.getLumberBaseProperties)(room.documentId, room.tenantId));
        promises.push(clientManager
            .removeClient(room.tenantId, room.documentId, clientId)
            .then(() => {
            // Keep track of disconnected clientIds so that we don't repeat the disconnect signal
            // for the same clientId if retrying when connectDocument completes after disconnectDocument.
            disconnectedClients.add(clientId);
        })
            .catch((error) => {
            server_services_telemetry_1.Lumberjack.error(`Failed to remove client ${clientId} from client manager`, (0, server_services_telemetry_1.getLumberBaseProperties)(room.documentId, room.tenantId), error);
        }));
        try {
            socket.emitToRoom((0, utils_2.getRoomId)(room), "signal", (0, utils_1.createRoomLeaveMessage)(clientId));
        }
        catch (error) {
            const errorMsg = `Failed to emit signal to room ${clientId}, ${(0, utils_2.getRoomId)(room)}.`;
            server_services_telemetry_1.Lumberjack.error(errorMsg, (0, server_services_telemetry_1.getLumberBaseProperties)(room.documentId, room.tenantId), error);
        }
        // Update session tracker upon disconnection
        if (collaborationSessionTracker) {
            const client = clientMap.get(clientId);
            const connectionTimestamp = connectionTimeMap.get(clientId);
            if (client) {
                collaborationSessionTracker
                    .endClientSession({
                    clientId,
                    joinedTime: connectionTimestamp ?? 0,
                    isSummarizerClient: (0, utils_2.isSummarizer)(client.details),
                    isWriteClient: (0, utils_2.isWriter)(client.scopes, client.mode),
                }, {
                    tenantId: room.tenantId,
                    documentId: room.documentId,
                })
                    .catch((error) => {
                    server_services_telemetry_1.Lumberjack.error("Failed to update collaboration session tracker for client disconnection", { messageMetaData }, error);
                });
            }
        }
    }
    return promises;
}
/**
 * Perform necessary cleanup when a client disconnects from a document.
 * @internal
 */
async function disconnectDocument(socket, nexusLambdaDependencies, nexusLambdaSettings, nexusLambdaConnectionStateTrackers) {
    // Clear token expiration timer on disconnection
    nexusLambdaConnectionStateTrackers.expirationTimer.clear();
    // Iterate over connection and room maps to disconnect and store connectivity time.
    const removeAndStoreP = [
        // Disconnect any orderer connections
        ...disconnectOrdererConnections(nexusLambdaDependencies, nexusLambdaSettings, nexusLambdaConnectionStateTrackers),
        // Send notification messages for all client IDs in the room map
        ...removeClientAndSendNotifications(socket, nexusLambdaDependencies, nexusLambdaConnectionStateTrackers),
    ];
    // Clear socket tracker upon disconnection
    if (nexusLambdaDependencies.socketTracker) {
        nexusLambdaDependencies.socketTracker.removeSocket(socket.id);
    }
    await Promise.all(removeAndStoreP);
}
exports.disconnectDocument = disconnectDocument;
//# sourceMappingURL=disconnect.js.map