"use strict";
/*!
 * Copyright (c) Microsoft Corporation and contributors. All rights reserved.
 * Licensed under the MIT License.
 */
var __createBinding = (this && this.__createBinding) || (Object.create ? (function(o, m, k, k2) {
    if (k2 === undefined) k2 = k;
    var desc = Object.getOwnPropertyDescriptor(m, k);
    if (!desc || ("get" in desc ? !m.__esModule : desc.writable || desc.configurable)) {
      desc = { enumerable: true, get: function() { return m[k]; } };
    }
    Object.defineProperty(o, k2, desc);
}) : (function(o, m, k, k2) {
    if (k2 === undefined) k2 = k;
    o[k2] = m[k];
}));
var __setModuleDefault = (this && this.__setModuleDefault) || (Object.create ? (function(o, v) {
    Object.defineProperty(o, "default", { enumerable: true, value: v });
}) : function(o, v) {
    o["default"] = v;
});
var __importStar = (this && this.__importStar) || function (mod) {
    if (mod && mod.__esModule) return mod;
    var result = {};
    if (mod != null) for (var k in mod) if (k !== "default" && Object.prototype.hasOwnProperty.call(mod, k)) __createBinding(result, mod, k);
    __setModuleDefault(result, mod);
    return result;
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.configureWebSocketServices = void 0;
const protocol_definitions_1 = require("@fluidframework/protocol-definitions");
const server_services_client_1 = require("@fluidframework/server-services-client");
const uuid_1 = require("uuid");
const core = __importStar(require("@fluidframework/server-services-core"));
const server_services_telemetry_1 = require("@fluidframework/server-services-telemetry");
const utils_1 = require("../utils");
const utils_2 = require("./utils");
const throttleAndUsage_1 = require("./throttleAndUsage");
const trace_1 = require("./trace");
const connect_1 = require("./connect");
const disconnect_1 = require("./disconnect");
const protocol_1 = require("./protocol");
// Sanitize the received op before sending.
function sanitizeMessage(message) {
    const sanitizedMessage = {
        clientSequenceNumber: message.clientSequenceNumber,
        contents: message.contents,
        metadata: message.metadata,
        referenceSequenceNumber: message.referenceSequenceNumber,
        traces: message.traces,
        type: message.type,
        compression: message.compression,
    };
    return sanitizedMessage;
}
/**
 * Converts a relayUserAgent string into a \<key,value\> map.
 * @param relayUserAgent - user agent string in the format "prop1:val1;prop2:val2;prop3:val3"
 */
function parseRelayUserAgent(relayUserAgent) {
    if (!relayUserAgent) {
        return {};
    }
    const map = {};
    const propertyKeyValuePairs = relayUserAgent
        .split(";")
        .map((keyValuePair) => keyValuePair.split(":"));
    for (const [key, value] of propertyKeyValuePairs) {
        // Trim key and value so that a user agent like "prop1:val1; prop2:val2" is parsed correctly.
        if (key && value)
            map[key.trim()] = value.trim();
    }
    return map;
}
// TODO: documentation
// eslint-disable-next-line jsdoc/require-description
/**
 * @internal
 */
function configureWebSocketServices(webSocketServer, ordererManager, tenantManager, storage, clientManager, metricLogger, logger, maxNumberOfClientsPerDocument = 1000000, numberOfMessagesPerTrace = 100, maxTokenLifetimeSec = 60 * 60, isTokenExpiryEnabled = false, isClientConnectivityCountingEnabled = false, isSignalUsageCountingEnabled = false, cache, connectThrottlerPerTenant, connectThrottlerPerCluster, submitOpThrottler, submitSignalThrottler, throttleAndUsageStorageManager, verifyMaxMessageSize, socketTracker, revokedTokenChecker, collaborationSessionEventEmitter, clusterDrainingChecker, collaborationSessionTracker, denyList) {
    const lambdaDependencies = {
        ordererManager,
        tenantManager,
        clientManager,
        logger,
        throttleAndUsageStorageManager,
        throttlers: {
            connectionsPerTenant: connectThrottlerPerTenant,
            connectionsPerCluster: connectThrottlerPerCluster,
            submitOps: submitOpThrottler,
            submitSignals: submitSignalThrottler,
        },
        socketTracker,
        revokedTokenChecker,
        clusterDrainingChecker,
        collaborationSessionEventEmitter,
        collaborationSessionTracker,
    };
    const lambdaSettings = {
        maxTokenLifetimeSec,
        isTokenExpiryEnabled,
        isClientConnectivityCountingEnabled,
        maxNumberOfClientsPerDocument,
        numberOfMessagesPerTrace,
    };
    webSocketServer.on("connection", (socket) => {
        // Timer to check token expiry for this socket connection
        const expirationTimer = new utils_2.ExpirationTimer(() => socket.disconnect(true));
        /**
         * Maps and sets to track various information related to client connections.
         * Note: These maps/sets are expected to have only one client id entry.
         */
        // Map from client IDs on this connection to the object ID and user info
        const connectionsMap = new Map();
        // Map from client IDs to room
        const roomMap = new Map();
        // Map from client Ids to scope
        const scopeMap = new Map();
        // Map from client Ids to client details
        const clientMap = new Map();
        // Map from client Ids to connection time.
        const connectionTimeMap = new Map();
        // Map from client Ids to supportedFeatures ()
        const supportedFeaturesMap = new Map();
        // Set of client Ids that have been disconnected from orderer.
        const disconnectedOrdererConnections = new Set();
        // Set of client Ids that have been disconnected from room and client manager.
        const disconnectedClients = new Set();
        const lambdaConnectionStateTrackers = {
            connectionsMap,
            roomMap,
            scopeMap,
            clientMap,
            connectionTimeMap,
            expirationTimer,
            disconnectedOrdererConnections,
            disconnectedClients,
            supportedFeaturesMap,
        };
        let connectDocumentComplete = false;
        let connectDocumentP;
        let disconnectDocumentP;
        const disposers = [socket.dispose?.bind(socket)];
        // Note connect is a reserved socket.io word so we use connect_document to represent the connect request
        // eslint-disable-next-line @typescript-eslint/no-misused-promises
        socket.on("connect_document", async (connectionMessage) => {
            if (!(0, protocol_1.isValidConnectionMessage)(connectionMessage)) {
                // If the connection message is invalid, emit an error and return.
                // This will prevent the connection from being established, but more importantly
                // it will provent the service from crashing due to an unhandled exception such as a type error.
                const error = new server_services_client_1.NetworkError(400, "Invalid connection message");
                // Attempt to log the connection message properties if they are available.
                // Be cautious to not log any sensitive information, such as message.token or message.user.
                const safeTelemetryProperties = typeof connectionMessage === "object" &&
                    connectionMessage !== null &&
                    typeof connectionMessage.id === "string" &&
                    typeof connectionMessage.tenantId === "string"
                    ? (0, server_services_telemetry_1.getLumberBaseProperties)(connectionMessage.id, connectionMessage.tenantId)
                    : {};
                if (typeof connectionMessage?.driverVersion === "string") {
                    safeTelemetryProperties.driverVersion = connectionMessage.driverVersion;
                }
                else if (typeof connectionMessage?.relayUserAgent === "string") {
                    safeTelemetryProperties.driverVersion = parseRelayUserAgent(connectionMessage.relayUserAgent).driverVersion;
                }
                server_services_telemetry_1.Lumberjack.warning("Received invalid connection message", safeTelemetryProperties, error);
                socket.emit("connect_document_error", error);
                return;
            }
            const userAgentInfo = parseRelayUserAgent(connectionMessage.relayUserAgent);
            const driverVersion = userAgentInfo.driverVersion;
            const baseLumberjackProperties = (0, server_services_telemetry_1.getLumberBaseProperties)(connectionMessage.id, connectionMessage.tenantId);
            const correlationId = (0, uuid_1.v4)();
            const properties = {
                ...baseLumberjackProperties,
                [server_services_telemetry_1.CommonProperties.clientDriverVersion]: driverVersion,
                [server_services_telemetry_1.CommonProperties.connectionCount]: connectionsMap.size,
                [server_services_telemetry_1.CommonProperties.connectionClients]: JSON.stringify([...connectionsMap.keys()]),
                [server_services_telemetry_1.CommonProperties.roomClients]: JSON.stringify([...roomMap.keys()]),
                [server_services_telemetry_1.BaseTelemetryProperties.correlationId]: correlationId,
            };
            connectDocumentP = (0, server_services_telemetry_1.getGlobalTelemetryContext)().bindPropertiesAsync({ correlationId, ...baseLumberjackProperties }, async () => (0, connect_1.connectDocument)(socket, lambdaDependencies, lambdaSettings, lambdaConnectionStateTrackers, connectionMessage, properties, denyList)
                .then((message) => {
                socket.emit("connect_document_success", message.connection);
                disposers.push(message.dispose);
            })
                .catch((error) => {
                socket.emit("connect_document_error", error);
                expirationTimer.clear();
            })
                .finally(() => {
                connectDocumentComplete = true;
                if (disconnectDocumentP) {
                    server_services_telemetry_1.Lumberjack.warning(`ConnectDocument completed after disconnect was handled.`);
                    // We have already received disconnect for this connection.
                    disconnectDocumentP
                        .catch((error) => {
                        server_services_telemetry_1.Lumberjack.error("Error encountered in disconnectDocumentP", { ...baseLumberjackProperties }, error);
                    })
                        .finally(() => {
                        // We might need to re-run disconnect handler after previous disconnect handler completes.
                        // DisconnectDocument internally handles the cases where we have already run disconnect for
                        // roomsMap and connectionsMap so that we don't duplicate disconnect efforts.
                        // The primary need for this retry is when we receive "disconnect" in the narrow window after
                        // "connect_document" but before "connectDocumentP" is defined.
                        const alreadyDisconnectedAllConnections = connectionsMap.size ===
                            disconnectedOrdererConnections.size;
                        const alreadyDisconnectedAllClients = roomMap.size === disconnectedClients.size;
                        if (alreadyDisconnectedAllConnections &&
                            alreadyDisconnectedAllClients) {
                            // Don't retry disconnect if all connections and clients are already handled.
                            return;
                        }
                        const disconnectRetryMetric = server_services_telemetry_1.Lumberjack.newLumberMetric(server_services_telemetry_1.LumberEventName.DisconnectDocumentRetry);
                        disconnectRetryMetric.setProperties({
                            ...baseLumberjackProperties,
                            [server_services_telemetry_1.CommonProperties.connectionCount]: connectionsMap.size,
                            [server_services_telemetry_1.CommonProperties.connectionClients]: JSON.stringify([
                                ...connectionsMap.keys(),
                            ]),
                            [server_services_telemetry_1.CommonProperties.roomClients]: JSON.stringify([
                                ...roomMap.keys(),
                            ]),
                        });
                        (0, disconnect_1.disconnectDocument)(socket, lambdaDependencies, lambdaSettings, lambdaConnectionStateTrackers)
                            .then(() => {
                            disconnectRetryMetric.success(`Successfully retried disconnect.`);
                        })
                            .catch((error) => {
                            disconnectRetryMetric.error(`Disconnect retry failed.`, error);
                        });
                    });
                }
            }));
        });
        // Message sent when a new operation is submitted to the router
        socket.on("submitOp", (clientId, messageBatches) => {
            // Verify the user has an orderer connection.
            const connection = connectionsMap.get(clientId);
            if (connection) {
                let messageCount = 0;
                for (const messageBatch of messageBatches) {
                    // Count all messages in each batch for accurate throttling calculation.
                    // Note: This is happening before message size checking. We won't process
                    // messages that are too large, so it is inaccurate to increment throttle
                    // counts for unprocessed messages.
                    messageCount += Array.isArray(messageBatch) ? messageBatch.length : 1;
                }
                const throttleError = (0, throttleAndUsage_1.checkThrottleAndUsage)(submitOpThrottler, (0, throttleAndUsage_1.getSubmitOpThrottleId)(clientId, connection.tenantId), connection.tenantId, logger, undefined, undefined, messageCount /* incrementWeight */);
                if (throttleError) {
                    const nackMessage = (0, utils_1.createNackMessage)(throttleError.code, protocol_definitions_1.NackErrorType.ThrottlingError, throttleError.message, throttleError.retryAfter);
                    socket.emit("nack", "", [nackMessage]);
                    return;
                }
                const lumberjackProperties = {
                    [server_services_telemetry_1.CommonProperties.clientId]: clientId,
                    ...(0, server_services_telemetry_1.getLumberBaseProperties)(connection.documentId, connection.tenantId),
                };
                const handleMessageBatchProcessingError = (error) => {
                    if ((0, server_services_client_1.isNetworkError)(error) && error.code === 413) {
                        server_services_telemetry_1.Lumberjack.info("Rejected too large operation(s)", lumberjackProperties);
                        socket.emit("nack", "", [
                            (0, utils_1.createNackMessage)(error.code, protocol_definitions_1.NackErrorType.BadRequestError, error.message),
                        ]);
                        return;
                    }
                    server_services_telemetry_1.Lumberjack.error("Error processing submitted op(s)", lumberjackProperties, error);
                };
                for (const messageBatch of messageBatches) {
                    const messages = Array.isArray(messageBatch)
                        ? messageBatch
                        : [messageBatch];
                    try {
                        const sanitized = messages.map((message) => {
                            if (verifyMaxMessageSize === true) {
                                // Local tests show `JSON.stringify` to be fast
                                // - <1ms for JSONs <100kb
                                // - ~2ms for JSONs ~256kb
                                // - ~6ms for JSONs ~1mb
                                // - ~38ms for JSONs ~10mb
                                // - ~344ms for JSONs ~100mb
                                // maxMessageSize is currently 16kb, so this check should be <1ms
                                const messageSize = JSON.stringify(message.contents).length;
                                const maxMessageSize = connection.serviceConfiguration.maxMessageSize;
                                if (messageSize > maxMessageSize) {
                                    server_services_telemetry_1.Lumberjack.error("Op size too large", {
                                        ...lumberjackProperties,
                                        messageSize,
                                        maxMessageSize,
                                    });
                                    // Exit early from processing message batch
                                    throw new server_services_client_1.NetworkError(413, "Op size too large");
                                }
                            }
                            const sanitizedMessage = sanitizeMessage(message);
                            const sanitizedMessageWithTrace = (0, trace_1.addNexusMessageTrace)(sanitizedMessage, numberOfMessagesPerTrace, connection.clientId, connection.tenantId, connection.documentId);
                            return sanitizedMessageWithTrace;
                        });
                        if (sanitized.length > 0) {
                            // Cannot await this order call without delaying other message batches in this submitOp.
                            connection
                                .order(sanitized)
                                .catch(handleMessageBatchProcessingError);
                        }
                    }
                    catch (error) {
                        handleMessageBatchProcessingError(error);
                    }
                }
            }
            else {
                let nackMessage;
                const clientScope = scopeMap.get(clientId);
                if (clientScope && (0, utils_2.hasWriteAccess)(clientScope)) {
                    nackMessage = (0, utils_1.createNackMessage)(400, protocol_definitions_1.NackErrorType.BadRequestError, "Readonly client");
                }
                else if (roomMap.has(clientId)) {
                    nackMessage = (0, utils_1.createNackMessage)(403, protocol_definitions_1.NackErrorType.InvalidScopeError, "Invalid scope");
                }
                else {
                    nackMessage = (0, utils_1.createNackMessage)(400, protocol_definitions_1.NackErrorType.BadRequestError, "Nonexistent client");
                }
                socket.emit("nack", "", [nackMessage]);
            }
        });
        // Message sent when a new signal is submitted to the router.
        socket.on("submitSignal", 
        // TODO: semantic documentation
        // eslint-disable-next-line jsdoc/require-description
        /**
         * @param contentBatches - typed as `unknown` array as it comes from wire and has not been validated.
         * v1 signals are expected to be an array of strings (Json.stringified `ISignalEnvelope`s from
         * [Container.submitSignal](https://github.com/microsoft/FluidFramework/blob/ccb26baf65be1cbe3f708ec0fe6887759c25be6d/packages/loader/container-loader/src/container.ts#L2292-L2294)
         * and sent via
         * [DocumentDeltaConnection.emitMessages](https://github.com/microsoft/FluidFramework/blob/ccb26baf65be1cbe3f708ec0fe6887759c25be6d/packages/drivers/driver-base/src/documentDeltaConnection.ts#L313C1-L321C4)),
         * but actual content is passed-thru and not decoded.
         *
         * v2 signals are expected to be an array of `ISentSignalMessage` objects.
         */
        (clientId, contentBatches) => {
            // Verify the user has subscription to the room.
            const room = roomMap.get(clientId);
            if (room) {
                if (!Array.isArray(contentBatches)) {
                    const nackMessage = (0, utils_1.createNackMessage)(400, protocol_definitions_1.NackErrorType.BadRequestError, "Invalid signal message");
                    socket.emit("nack", "", [nackMessage]);
                    return;
                }
                let messageCount = 0;
                for (const contentBatch of contentBatches) {
                    // Count all messages in each batch for accurate throttling calculation.
                    messageCount += Array.isArray(contentBatch) ? contentBatch.length : 1;
                }
                const signalUsageData = {
                    value: 0,
                    tenantId: room.tenantId,
                    documentId: room.documentId,
                    clientId,
                };
                const throttleError = (0, throttleAndUsage_1.checkThrottleAndUsage)(submitSignalThrottler, (0, throttleAndUsage_1.getSubmitSignalThrottleId)(clientId, room.tenantId), room.tenantId, logger, isSignalUsageCountingEnabled ? core.signalUsageStorageId : undefined, isSignalUsageCountingEnabled ? signalUsageData : undefined, messageCount /* incrementWeight */);
                if (throttleError) {
                    const nackMessage = (0, utils_1.createNackMessage)(throttleError.code, protocol_definitions_1.NackErrorType.ThrottlingError, throttleError.message, throttleError.retryAfter);
                    socket.emit("nack", "", [nackMessage]);
                    return;
                }
                if (supportedFeaturesMap.get(clientId)?.submit_signals_v2) {
                    for (const signal of contentBatches) {
                        if ((0, utils_2.isSentSignalMessage)(signal)) {
                            const signalMessage = {
                                ...signal,
                                clientId,
                            };
                            const roomId = signal.targetClientId === undefined
                                ? (0, utils_2.getRoomId)(room)
                                : (0, utils_2.getClientSpecificRoomId)(signal.targetClientId);
                            socket.emitToRoom(roomId, "signal", signalMessage);
                        }
                        else {
                            // If the signal is not in the expected format, nack the message.
                            // This will disconnect client from the socket.
                            // No signals sent after this message will be processed.
                            const nackMessage = (0, utils_1.createNackMessage)(400, protocol_definitions_1.NackErrorType.BadRequestError, "Invalid signal message");
                            socket.emit("nack", "", [nackMessage]);
                            return;
                        }
                    }
                }
                else {
                    for (const contentBatch of contentBatches) {
                        const contents = Array.isArray(contentBatch)
                            ? contentBatch
                            : [contentBatch];
                        for (const content of contents) {
                            const signalMessage = {
                                clientId,
                                content,
                            };
                            const roomId = (0, utils_2.getRoomId)(room);
                            socket.emitToRoom(roomId, "signal", signalMessage);
                        }
                    }
                }
            }
            else {
                const nackMessage = (0, utils_1.createNackMessage)(400, protocol_definitions_1.NackErrorType.BadRequestError, "Nonexistent client");
                socket.emit("nack", "", [nackMessage]);
            }
        });
        // eslint-disable-next-line @typescript-eslint/no-misused-promises
        socket.on("disconnect", async (reason) => {
            if (!connectDocumentComplete && connectDocumentP) {
                server_services_telemetry_1.Lumberjack.warning(`Socket connection disconnected before ConnectDocument completed.`);
                // Wait for document connection to finish before disconnecting.
                // If disconnect fires before roomMap or connectionsMap are updated, we can be left with
                // hanging connections and clients.
                await connectDocumentP.catch(() => { });
            }
            const disconnectMetric = server_services_telemetry_1.Lumberjack.newLumberMetric(server_services_telemetry_1.LumberEventName.DisconnectDocument);
            disconnectMetric.setProperties({
                [server_services_telemetry_1.CommonProperties.connectionCount]: connectionsMap.size,
                [server_services_telemetry_1.CommonProperties.connectionClients]: JSON.stringify([...connectionsMap.keys()]),
                [server_services_telemetry_1.CommonProperties.roomClients]: JSON.stringify([...roomMap.keys()]),
                // Socket.io provides disconnect reason as a string. If it is not a string, it might not be a socket.io socket, so don't log anything.
                // A list of possible reasons can be found here: https://socket.io/docs/v4/server-socket-instance/#disconnect
                [server_services_telemetry_1.CommonProperties.disconnectReason]: typeof reason === "string" ? reason : undefined,
            });
            if (roomMap.size > 0) {
                const rooms = [...roomMap.values()];
                const documentId = rooms[0].documentId;
                const tenantId = rooms[0].tenantId;
                disconnectMetric.setProperties({
                    ...(0, server_services_telemetry_1.getLumberBaseProperties)(documentId, tenantId),
                });
            }
            try {
                disconnectDocumentP = (0, disconnect_1.disconnectDocument)(socket, lambdaDependencies, lambdaSettings, lambdaConnectionStateTrackers);
                await disconnectDocumentP;
                disconnectMetric.success(`Successfully disconnected.`);
            }
            catch (error) {
                disconnectMetric.error(`Disconnect failed.`, error);
            }
            // Dispose all resources and clear list.
            for (const dispose of disposers) {
                if (dispose) {
                    dispose();
                }
            }
            disposers.splice(0, disposers.length);
        });
    });
}
exports.configureWebSocketServices = configureWebSocketServices;
//# sourceMappingURL=index.js.map